#include <db.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>

#define BULK_FETCH_BUFFER_SIZE (8*1024)

int cread(DB *dbp) {
    DBC *dbcp;
    DBT key, data;
    int ret;
    char buf[100];
    char buffer[BULK_FETCH_BUFFER_SIZE];

    /*
     * The DB handle for a Btree database supporting duplicate data
     * items is the argument; acquire a cursor for the database.
     */


    /* Initialize our DBTs. */
    memset(&key, 0, sizeof(DBT));
    memset(&data, 0, sizeof(DBT));

    key.data = "new key";
    key.size = strlen(key.data) + 1;

    data.ulen = sizeof(buffer);
    data.size = sizeof(buffer);
    data.data = buffer;
    data.flags = DB_DBT_USERMEM;

    /* Do forever to force a deadlock */
    while(1) {
        if ((ret = dbp->cursor(dbp, NULL, &dbcp, 0)) != 0) {
            dbp->err(dbp, ret, "DB->cursor");
        }

        if ((ret = dbcp->get(dbcp, &key, &data, DB_SET|DB_MULTIPLE)) != 0)
            dbp->err(dbp, ret, "cursor->get");
        while ((ret = dbcp->get(dbcp, &key, &data, DB_NEXT_DUP|DB_MULTIPLE)) == 0) {
        }
        if (ret != DB_NOTFOUND) {
            dbp->err(dbp, ret, "cursor->get");
        }

        if ((ret = dbcp->close(dbcp)) != 0)
            dbp->err(dbp, ret, "DBcursor->close");
    }
    return (0);
}

/* Could be put somewhere reusable but copying is easier /shrug */
int main(int argc, char *argv[]) {
    DB_ENV *myEnv;            /* Env structure handle */
    DB *dbp;                  /* DB structure handle */
    u_int32_t db_flags;       /* database open flags */
    u_int32_t env_flags;      /* env open flags */
    int ret;                  /* function return value */

    /*
       Create an environment object and initialize it for error
       reporting.
    */
    ret = db_env_create(&myEnv, 0);
    if (ret != 0) {
        fprintf(stderr, "Error creating env handle: %s\n", db_strerror(ret));
        return -1;
    }

    /* Open the environment. */
    env_flags = DB_CREATE |    /* If the environment does not exist,
                                * create it. */
                DB_INIT_MPOOL; /* Initialize the in-memory cache. */
    env_flags |= (DB_INIT_TXN | DB_INIT_LOG | DB_INIT_LOCK);


    ret = myEnv->open(myEnv,   /* DB_ENV ptr */
      "/tmp/opddeadlock/testEnv",      /* env home directory */
      env_flags,               /* Open flags */
      0);                      /* File mode (default) */
    if (ret != 0) {
        fprintf(stderr, "Environment open failed: %s", db_strerror(ret));
        return -1;
    }

    /*
     * Initialize the DB structure. Pass the pointer
     * to the environment in which this DB is opened.
     */
    ret = db_create(&dbp, myEnv, 0);
    if (ret != 0) {
      /* Error handling goes here */
        dbp->err(dbp, ret, "DB->create");
        goto err;
    }

    ret = dbp->set_flags(dbp, DB_DUP|DB_DUPSORT);
    if (ret != 0) {
        /* Error handling goes here */
        dbp->err(dbp, ret, "DB->set_flags");
        goto err;
    }
    dbp->set_pagesize(dbp, 512);


    /* Database open flags */
    db_flags = DB_CREATE;    /* If the database does not exist,
                              * create it.*/

    /* open the database */
    ret = dbp->open(dbp,        /* DB structure pointer */
                    NULL,       /* Transaction pointer */
                    "my_db.db", /* On-disk file that holds the database. */
                    NULL,       /* Optional logical database name */
                    DB_BTREE,   /* Database access method */
                    db_flags,   /* Open flags */
                    0);         /* File mode (using defaults) */

    cread(dbp);

err:
    /*
    * Close the database and environment
    */

    if (dbp != NULL) {
        dbp->close(dbp, 0);
    }

    if (myEnv != NULL) {
        myEnv->close(myEnv, 0);
    }
}
