#include <db.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/time.h>

static char *rand_string(char *str, size_t size)
{
    const char charset[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
    if (size) {
        --size;
        for (size_t n = 0; n < size; n++) {
            int key = rand() % (int) (sizeof charset - 1);
            str[n] = charset[key];
        }
        str[size] = '\0';
    }
    return str;
}

int store(DB *dbp) {
    DBC *dbcp;
    DBT key, data;
    int ret;
    char buf[100];

    struct timeval t1, t2;
    int elapsed;

    /*
     * The DB handle for a Btree database supporting duplicate data
     * items is the argument; acquire a cursor for the database.
     */
    if ((ret = dbp->cursor(dbp, NULL, &dbcp, 0)) != 0) {
        dbp->err(dbp, ret, "DB->cursor");
        goto err;
    }

    /* Initialize the key. */
    memset(&key, 0, sizeof(key));
    key.data = "new key";
    key.size = strlen(key.data) + 1;
    gettimeofday(&t1, NULL);

    while(1) {
        memset(&data, 0, sizeof(data));
        /*
         * Create a pseudo-random string and hope we deadlock sooner
         * than a duplicate data item is hit.
         */
        data.data = rand_string(buf, 100);
        // printf("%s\n", data.data);
        data.size = strlen(data.data) + 1;

        if ((ret = dbp->put(dbp, NULL, &key, &data, 0)) != 0) {
            dbp->err(dbp, ret, "DB->put");
            goto err;
        }
        gettimeofday(&t2, NULL);
        elapsed = (t2.tv_sec - t1.tv_sec) * 1000;
        elapsed += (t2.tv_usec - t1.tv_usec) / 1000.0;
        //printf("%d\n", elapsed);
        if (elapsed > 30000) //30s
            exit(0);
    }

err:    if ((ret = dbcp->close(dbcp)) != 0)
        dbp->err(dbp, ret, "DBcursor->close");

    return (0);
}

/* Could be put somewhere reusable but copying is easier /shrug */
int main(int argc, char *argv[]) {
    DB_ENV *myEnv;            /* Env structure handle */
    DB *dbp;                  /* DB structure handle */
    u_int32_t db_flags;       /* database open flags */
    u_int32_t env_flags;      /* env open flags */
    int ret;                  /* function return value */

    /*
       Create an environment object and initialize it for error
       reporting.
    */
    ret = db_env_create(&myEnv, 0);
    if (ret != 0) {
        fprintf(stderr, "Error creating env handle: %s\n", db_strerror(ret));
        return -1;
    }

    /* Open the environment. */
    env_flags = DB_CREATE |    /* If the environment does not exist,
                                * create it. */
                DB_INIT_MPOOL; /* Initialize the in-memory cache. */
    env_flags |= (DB_INIT_TXN | DB_INIT_LOG | DB_INIT_LOCK);


    ret = myEnv->open(myEnv,   /* DB_ENV ptr */
      "/tmp/opddeadlock/testEnv",      /* env home directory */
      env_flags,               /* Open flags */
      0);                      /* File mode (default) */
    if (ret != 0) {
        fprintf(stderr, "Environment open failed: %s", db_strerror(ret));
        return -1;
    }

    /*
     * Initialize the DB structure. Pass the pointer
     * to the environment in which this DB is opened.
     */
    ret = db_create(&dbp, myEnv, 0);
    if (ret != 0) {
      /* Error handling goes here */
        dbp->err(dbp, ret, "DB->create");
        goto err;
    }

    ret = dbp->set_flags(dbp, DB_DUP|DB_DUPSORT);
    if (ret != 0) {
        /* Error handling goes here */
        dbp->err(dbp, ret, "DB->set_flags");
        goto err;
    }
    dbp->set_pagesize(dbp, 512);


    /* Database open flags */
    db_flags = DB_CREATE;    /* If the database does not exist,
                              * create it.*/

    /* open the database */
    ret = dbp->open(dbp,        /* DB structure pointer */
                    NULL,       /* Transaction pointer */
                    "my_db.db", /* On-disk file that holds the database. */
                    NULL,       /* Optional logical database name */
                    DB_BTREE,   /* Database access method */
                    db_flags,   /* Open flags */
                    0);         /* File mode (using defaults) */

    store(dbp);

err:
    /*
    * Close the database and environment
    */

    if (dbp != NULL) {
        dbp->close(dbp, 0);
    }

    if (myEnv != NULL) {
        myEnv->close(myEnv, 0);
    }
}
